'use strict';
Object.defineProperty(exports, '__esModule', { value: true });
exports.IncrementalGraph = void 0;
const BoxedPromiseOrValue_js_1 = require('../jsutils/BoxedPromiseOrValue.js');
const invariant_js_1 = require('../jsutils/invariant.js');
const isPromise_js_1 = require('../jsutils/isPromise.js');
const promiseWithResolvers_js_1 = require('../jsutils/promiseWithResolvers.js');
const DeferredFragments_js_1 = require('./DeferredFragments.js');
const types_js_1 = require('./types.js');
/**
 * @internal
 */
class IncrementalGraph {
  constructor() {
    this._deferredFragmentFactory =
      new DeferredFragments_js_1.DeferredFragmentFactory();
    this._rootNodes = new Set();
    this._completedQueue = [];
    this._nextQueue = [];
  }
  getNewRootNodes(incrementalDataRecords) {
    const initialResultChildren = new Set();
    this._addIncrementalDataRecords(
      incrementalDataRecords,
      undefined,
      initialResultChildren,
    );
    return this._promoteNonEmptyToRoot(initialResultChildren);
  }
  addCompletedSuccessfulExecutionGroup(successfulExecutionGroup) {
    const { pendingExecutionGroup, incrementalDataRecords } =
      successfulExecutionGroup;
    const { deferUsages, path } = pendingExecutionGroup;
    const deferredFragmentRecords = [];
    for (const deferUsage of deferUsages) {
      const deferredFragmentRecord = this._deferredFragmentFactory.get(
        deferUsage,
        path,
      );
      deferredFragmentRecords.push(deferredFragmentRecord);
      const { pendingExecutionGroups, successfulExecutionGroups } =
        deferredFragmentRecord;
      pendingExecutionGroups.delete(pendingExecutionGroup);
      successfulExecutionGroups.add(successfulExecutionGroup);
    }
    if (incrementalDataRecords !== undefined) {
      this._addIncrementalDataRecords(
        incrementalDataRecords,
        deferredFragmentRecords,
      );
    }
  }
  getDeepestDeferredFragmentAtRoot(initialDeferUsage, deferUsages, path) {
    let bestDeferUsage = initialDeferUsage;
    let maxDepth = initialDeferUsage.depth;
    for (const deferUsage of deferUsages) {
      if (deferUsage === initialDeferUsage) {
        continue;
      }
      const depth = deferUsage.depth;
      if (depth > maxDepth) {
        maxDepth = depth;
        bestDeferUsage = deferUsage;
      }
    }
    return this._deferredFragmentFactory.get(bestDeferUsage, path);
  }
  *currentCompletedBatch() {
    let completed;
    while ((completed = this._completedQueue.shift()) !== undefined) {
      yield completed;
    }
    if (this._rootNodes.size === 0) {
      for (const resolve of this._nextQueue) {
        resolve(undefined);
      }
    }
  }
  nextCompletedBatch() {
    const { promise, resolve } = (0,
    promiseWithResolvers_js_1.promiseWithResolvers)();
    this._nextQueue.push(resolve);
    return promise;
  }
  abort() {
    for (const resolve of this._nextQueue) {
      resolve(undefined);
    }
  }
  hasNext() {
    return this._rootNodes.size > 0;
  }
  completeDeferredFragment(deferUsage, path) {
    const deferredFragmentRecord = this._deferredFragmentFactory.get(
      deferUsage,
      path,
    );
    if (
      !this._rootNodes.has(deferredFragmentRecord) ||
      deferredFragmentRecord.pendingExecutionGroups.size > 0
    ) {
      return;
    }
    const successfulExecutionGroups = Array.from(
      deferredFragmentRecord.successfulExecutionGroups,
    );
    this._rootNodes.delete(deferredFragmentRecord);
    for (const successfulExecutionGroup of successfulExecutionGroups) {
      const { deferUsages, path: resultPath } =
        successfulExecutionGroup.pendingExecutionGroup;
      for (const otherDeferUsage of deferUsages) {
        const otherDeferredFragmentRecord = this._deferredFragmentFactory.get(
          otherDeferUsage,
          resultPath,
        );
        otherDeferredFragmentRecord.successfulExecutionGroups.delete(
          successfulExecutionGroup,
        );
      }
    }
    const newRootNodes = this._promoteNonEmptyToRoot(
      deferredFragmentRecord.children,
    );
    return { deferredFragmentRecord, newRootNodes, successfulExecutionGroups };
  }
  removeDeferredFragment(deferUsage, path) {
    const deferredFragmentRecord = this._deferredFragmentFactory.get(
      deferUsage,
      path,
    );
    if (!this._rootNodes.has(deferredFragmentRecord)) {
      return;
    }
    this._rootNodes.delete(deferredFragmentRecord);
    return deferredFragmentRecord;
  }
  removeStream(streamRecord) {
    this._rootNodes.delete(streamRecord);
  }
  _addIncrementalDataRecords(
    incrementalDataRecords,
    parents,
    initialResultChildren,
  ) {
    for (const incrementalDataRecord of incrementalDataRecords) {
      if ((0, types_js_1.isPendingExecutionGroup)(incrementalDataRecord)) {
        const { deferUsages, path } = incrementalDataRecord;
        for (const deferUsage of deferUsages) {
          const deferredFragmentRecord = this._deferredFragmentFactory.get(
            deferUsage,
            path,
          );
          this._addDeferredFragment(
            deferredFragmentRecord,
            initialResultChildren,
          );
          deferredFragmentRecord.pendingExecutionGroups.add(
            incrementalDataRecord,
          );
        }
        if (this._completesRootNode(incrementalDataRecord)) {
          this._onExecutionGroup(incrementalDataRecord);
        }
      } else if (parents === undefined) {
        initialResultChildren !== undefined ||
          (0, invariant_js_1.invariant)(false);
        initialResultChildren.add(incrementalDataRecord);
      } else {
        for (const parent of parents) {
          this._addDeferredFragment(parent, initialResultChildren);
          parent.children.add(incrementalDataRecord);
        }
      }
    }
  }
  _promoteNonEmptyToRoot(maybeEmptyNewRootNodes) {
    const newRootNodes = [];
    for (const node of maybeEmptyNewRootNodes) {
      if ((0, DeferredFragments_js_1.isDeferredFragmentRecord)(node)) {
        if (node.pendingExecutionGroups.size > 0) {
          for (const pendingExecutionGroup of node.pendingExecutionGroups) {
            if (!this._completesRootNode(pendingExecutionGroup)) {
              this._onExecutionGroup(pendingExecutionGroup);
            }
          }
          this._rootNodes.add(node);
          newRootNodes.push(node);
          continue;
        }
        for (const child of node.children) {
          maybeEmptyNewRootNodes.add(child);
        }
      } else {
        this._rootNodes.add(node);
        newRootNodes.push(node);
        // eslint-disable-next-line @typescript-eslint/no-floating-promises
        this._onStreamItems(node);
      }
    }
    return newRootNodes;
  }
  _completesRootNode(pendingExecutionGroup) {
    const { deferUsages, path } = pendingExecutionGroup;
    for (const deferUsage of deferUsages) {
      const deferredFragmentRecord = this._deferredFragmentFactory.get(
        deferUsage,
        path,
      );
      if (this._rootNodes.has(deferredFragmentRecord)) {
        return true;
      }
    }
    return false;
  }
  _addDeferredFragment(deferredFragmentRecord, initialResultChildren) {
    if (this._rootNodes.has(deferredFragmentRecord)) {
      return;
    }
    const parentDeferUsage = deferredFragmentRecord.parentDeferUsage;
    if (parentDeferUsage === undefined) {
      initialResultChildren !== undefined ||
        (0, invariant_js_1.invariant)(false);
      initialResultChildren.add(deferredFragmentRecord);
      return;
    }
    const parent = this._deferredFragmentFactory.get(
      parentDeferUsage,
      deferredFragmentRecord.path,
    );
    parent.children.add(deferredFragmentRecord);
    this._addDeferredFragment(parent, initialResultChildren);
  }
  _onExecutionGroup(pendingExecutionGroup) {
    let completedExecutionGroup = pendingExecutionGroup.result;
    if (
      !(
        completedExecutionGroup instanceof
        BoxedPromiseOrValue_js_1.BoxedPromiseOrValue
      )
    ) {
      completedExecutionGroup = completedExecutionGroup();
    }
    const value = completedExecutionGroup.value;
    if ((0, isPromise_js_1.isPromise)(value)) {
      // eslint-disable-next-line @typescript-eslint/no-floating-promises
      value.then((resolved) => this._enqueue(resolved));
    } else {
      this._enqueue(value);
    }
  }
  async _onStreamItems(streamRecord) {
    let items = [];
    let errors = [];
    let incrementalDataRecords = [];
    const streamItemQueue = streamRecord.streamItemQueue;
    let streamItemRecord;
    while ((streamItemRecord = streamItemQueue.shift()) !== undefined) {
      let result =
        streamItemRecord instanceof BoxedPromiseOrValue_js_1.BoxedPromiseOrValue
          ? streamItemRecord.value
          : streamItemRecord().value;
      if ((0, isPromise_js_1.isPromise)(result)) {
        if (items.length > 0) {
          this._enqueue({
            streamRecord,
            result:
              // TODO add additional test case or rework for coverage
              errors.length > 0 /* c8 ignore start */
                ? { items, errors } /* c8 ignore stop */
                : { items },
            incrementalDataRecords,
          });
          items = [];
          errors = [];
          incrementalDataRecords = [];
        }
        // eslint-disable-next-line no-await-in-loop
        result = await result;
        // wait an additional tick to coalesce resolving additional promises
        // within the queue
        // eslint-disable-next-line no-await-in-loop
        await Promise.resolve();
      }
      if (result.item === undefined) {
        if (items.length > 0) {
          this._enqueue({
            streamRecord,
            result: errors.length > 0 ? { items, errors } : { items },
            incrementalDataRecords,
          });
        }
        this._enqueue(
          result.errors === undefined
            ? { streamRecord }
            : {
                streamRecord,
                errors: result.errors,
              },
        );
        return;
      }
      items.push(result.item);
      if (result.errors !== undefined) {
        errors.push(...result.errors);
      }
      if (result.incrementalDataRecords !== undefined) {
        incrementalDataRecords.push(...result.incrementalDataRecords);
      }
    }
  }
  _enqueue(completed) {
    this._completedQueue.push(completed);
    const next = this._nextQueue.shift();
    if (next === undefined) {
      return;
    }
    next(this.currentCompletedBatch());
  }
}
exports.IncrementalGraph = IncrementalGraph;
